from strenum import StrEnum


class HTTPMethods(StrEnum):
    """Available HTTP methods."""

    POST = "POST"
    GET = "GET"


class Networks(StrEnum):
    """Cryptobot networks"""

    MAIN_NET = "https://pay.crypt.bot"
    TEST_NET = "https://testnet-pay.crypt.bot"


class Assets(StrEnum):
    """Cryptobot assets"""

    BTC = "BTC"
    TON = "TON"
    ETH = "ETH"
    USDT = "USDT"
    USDC = "USDC"
    BNB = "BNB"
    TRX = "TRX"
    LTC = "LTC"
    GRAM = "GRAM"
    NOT = "NOT"
    MY = "MY"
    SOL = "SOL"

    @classmethod
    def values(cls):
        return list(map(lambda asset: asset.value, cls))


class Fiat(StrEnum):
    RUB = "RUB"
    USD = "USD"
    EUR = "EUR"
    BYN = "BYN"
    UAH = "UAH"
    KZT = "KZT"
    UZS = "UZS"
    GEL = "GEL"
    TRY = "TRY"
    AMD = "AMD"
    THB = "THB"
    INR = "INR"
    BRL = "BRL"
    IDR = "IDR"
    AZN = "AZN"
    AED = "AED"
    PLN = "PLN"
    ILS = "ILS"
    KGS = "KGS"
    TJS = "TJS"

    @classmethod
    def values(cls):
        return list(map(lambda fiat: fiat.value, cls))


class PaidButtons(StrEnum):
    """Cryptobot paid button names"""

    VIEW_ITEM = "viewItem"
    OPEN_CHANNEL = "openChannel"
    OPEN_BOT = "openBot"
    CALLBACK = "callback"


class InvoiceStatus(StrEnum):
    """Invoice status"""

    ACTIVE = "active"
    PAID = "paid"
    EXPIRED = "expired"


class CheckStatus(StrEnum):
    """Check status"""

    ACTIVE = "active"
    ACTIVATED = "activated"


class CurrencyType(StrEnum):
    """Currency type"""

    CRYPTO = "crypto"
    FIAT = "fiat"



from pydantic import BaseModel

from typing import Optional


class Currency(BaseModel):
    is_blockchain: bool
    is_stablecoin: bool
    is_fiat: bool
    name: str
    code: str
    url: Optional[str] = None
    decimals: int



import asyncio
import ssl
from typing import Optional

import certifi
from aiohttp import ClientSession, TCPConnector
from aiohttp.typedefs import StrOrURL

from .exceptions import CryptoPayAPIError


class BaseClient:
    """Base aiohttp client"""

    def __init__(self) -> None:
        """
        Set defaults on object init.
            By default `self._session` is None.
            It will be created on a first API request.
            The second request will use the same `self._session`.
        """
        self._loop = asyncio.get_event_loop()
        self._session: Optional[ClientSession] = None

    def get_session(self, **kwargs):
        """Get cached session. One session per instance."""
        if isinstance(self._session, ClientSession) and not self._session.closed:
            return self._session

        ssl_context = ssl.create_default_context(cafile=certifi.where())
        connector = TCPConnector(ssl=ssl_context)

        self._session = ClientSession(connector=connector, **kwargs)
        return self._session

    async def _make_request(self, method: str, url: StrOrURL, **kwargs) -> dict:
        """
        Make a request.
            :param method: HTTP Method
            :param url: endpoint link
            :param kwargs: data, params, json and other...
            :return: status and result or exception
        """
        session = self.get_session()

        async with session.request(method, url, **kwargs) as response:
            response = await response.json(content_type="application/json")
        return self._validate_response(response)

    @staticmethod
    def _validate_response(response: dict) -> dict:
        """Validate response"""
        if not response.get("ok"):
            name = response["error"]["name"]
            code = response["error"]["code"]
            raise CryptoPayAPIError(code, name)
        return response

    async def close(self):
        """Close the session graceful."""
        if not isinstance(self._session, ClientSession):
            return

        if self._session.closed:
            return

        await self._session.close()



from datetime import datetime

from pydantic import BaseModel
from typing import Union


class AppStats(BaseModel):
    volume: Union[int, float]
    conversion: Union[int, float]
    unique_users_count: int
    created_invoice_count: int
    paid_invoice_count: int
    start_at: datetime
    end_at: datetime



import gc
from typing import Optional, Type, Union


class CodeErrorFactory(Exception):
    """CryptoPay API Exception"""

    def __init__(self, code: int = None, name: str = None) -> None:
        self.code = int(code) if code else None
        self.name = name
        super().__init__(self.code)

    @classmethod
    def __call__(
        cls, code: Optional[int] = None, name: Optional[int] = None
    ) -> Union["CodeErrorFactory", Type["CodeErrorFactory"]]:
        if name:
            return cls.exception_to_raise(code, str(name))
        return cls.exception_to_handle(code)

    @classmethod
    def exception_to_handle(
        cls, code: Optional[int] = None
    ) -> Type["CodeErrorFactory"]:
        if code is None:
            return cls

        catch_exc_classname = cls.generate_exc_classname(code)

        for obj in gc.get_objects():
            if obj.__class__.__name__ == catch_exc_classname:
                return obj.__class__

        return type(catch_exc_classname, (cls,), {})

    @classmethod
    def exception_to_raise(cls, code: int, name: str) -> "CodeErrorFactory":
        """Returns an error with error code and error_name"""
        exception_type = type(cls.generate_exc_classname(code), (cls,), {})
        return exception_type(code, name)

    @classmethod
    def generate_exc_classname(cls, code: Optional[int]) -> str:
        """Generates unique exception classname based on error code"""
        return f"{cls.__name__}_{code}"

    def __str__(self):
        return f"[{self.code}] {self.name}\n"



from pydantic import BaseModel

from typing import Union, Optional
from datetime import datetime

from ..const import Assets


class Transfer(BaseModel):
    transfer_id: int
    user_id: int
    asset: Union[Assets, str]
    amount: Union[int, float]
    status: str
    completed_at: datetime
    comment: Optional[str] = None



from pydantic import BaseModel
from datetime import datetime

from .invoice import Invoice


class Update(BaseModel):
    update_id: int
    update_type: str
    request_date: datetime
    payload: Invoice



from pydantic import BaseModel

from typing import Union, Optional
from datetime import datetime

from ..const import Assets, CheckStatus


class Check(BaseModel):
    check_id: int
    hash: str
    asset: Union[Assets, str]
    amount: Union[int, float]
    bot_check_url: str
    status: Union[CheckStatus, str]
    created_at: datetime
    activated_at: Optional[datetime] = None
